<?php

/**
 * @author Davit Araqelyan <davaraqelyan111@gmail.com>
 */

namespace base\payments;

use base\interfaces\IPaydateCalculator;

class PaydateCalculator implements IPaydateCalculator {

    /**
     * Holiday list
     * @var array
     */
    private $_holidays = array();

    /**
     * Payment Method list and methods info
     * @var array
     */
    private $_paymentMethods = array(
        'weekly' => array(
            'count' => 1,
            'unit' => 'week',
        ),
        'monthly' => array(
            'count' => 1,
            'unit' => 'month',
        ),
        'biweekly' => array(
            'count' => 2,
            'unit' => 'week',
        ),
    );

    /**
     * Payment method name
     * @var string 
     */
    private $_method;

    /**
     * Real payment date , which isn't changed by rules
     * @var string
     */
    private $_realPaymentDate;

    public function calculateNextPaydates($paydate_model, $paydate_one, $paydate_two, $number_of_paydates = 10) {
        $dateList = array();
       
        if ($this->isValidPaydate($paydate_one) && $this->isValidPaydate($paydate_two)) {
            $this->initRealPaymentDate($paydate_model, $paydate_one, $paydate_two);
            $today = new \DateTime();
            $realPaymentDay = new \DateTime($this->_realPaymentDate);
            
            $unit = $this->_paymentMethods[$this->_method]['unit'];
            $count = $this->_paymentMethods[$this->_method]['count'];
            if ($realPaymentDay > $today) {
                $startDate = $realPaymentDay->format('Y-m-d');
                while ($realPaymentDay->modify('-' . $count . ' ' . $unit) > $today) {
                    $startDate = $realPaymentDay->format('Y-m-d');
                    $startDate = $this->nearestValidPaymentDate($startDate);
                }
            } else {
                if ($realPaymentDay <= $today) {
                    while ($realPaymentDay <= $today) {
                        $realPaymentDay->modify('+' . $count . ' ' . $unit);
                        $startDate = $realPaymentDay->format('Y-m-d');
                        $startDate = $this->nearestValidPaymentDate($startDate);
                    }
                }
            }
            $dateList = $this->getNextPaydatesFrom($startDate, $number_of_paydates);
        }else{
            throw new \Exception('One of given dates weekend or holiday');
        }
        return $dateList;
    }

    /**
     * Return list of next payment dates form given start date
     * @param string $startDate 
     * @param integer $count count of list
     * @return array
     */
    public function getNextPaydatesFrom($startDate, $count = 10) {
        $dateList = array($startDate);
        $dateTime = new \DateTime($startDate);
        $unit = $this->_paymentMethods[$this->_method]['unit'];
        $countUnits = $this->_paymentMethods[$this->_method]['count'];
        for ($i = 0; $i < $count - 1; $i++) {
            $dateTime->modify('+' . $countUnits . ' ' . $unit);
            $date = $dateTime->format('Y-m-d');
            
            $dateList[] = $this->nearestValidPaymentDate($date);
        }
        $today = new \DateTime();
        if(in_array($today->format('Y-m-d'), $dateList)){
            array_shift($dateList);
            $dateTime->modify('+' . $countUnits . ' ' . $unit);
            $date = $dateTime->format('Y-m-d');
            
            $dateList[] = $this->nearestValidPaymentDate($date);
        }
        return $dateList;
    }

    public function increaseDate($date, $count, $unit = 'days') {
        $dateTime = new \DateTime($date);
        $dateTime->modify('+' . $count . ' ' . $unit);
        return $dateTime->format('Y-m-d');
    }

    public function decreaseDate($date, $count, $unit = 'days') {
        $dateTime = new \DateTime($date);
        $dateTime->modify('-' . $count . ' ' . $unit);
        return $dateTime->format('Y-m-d');
    }

    public function isHoliday($date) {
        return in_array($date, $this->_holidays);
    }

    public function isValidPaydate($date) {
        $pattern = '/^\d{4}\-\d{2}\-\d{2}$/';
        if (!preg_match($pattern, $date)) {
            return false;
        }
        if (!$this->isHoliday($date) && !$this->isWeekend($date)) {
            return true;
        } else {
            return false;
        }
    }

    public function isWeekend($date) {
        return (date('N', strtotime($date)) >= 6);
    }

    /**
     * Holidays list (fomat dd/mm/YYYY)
     * @param array $holidays
     */
    public function setHolidays($holidays) {
        foreach ($holidays as $holiday) {
            $dt = new \DateTime($holiday);
            $this->_holidays[] = $dt->format('Y-m-d');
        }
    }

    /**
     * Finding real payemnt date from given two payemnt dates examples
     * @param string $method Payment method (MONTHLY, BIWEEKLY, WEEKLY)
     * @param string $paydate_one Example of payment date (format Y-m-d)
     * @param string $paydate_two Example of payment date (format Y-m-d) 
     */
    public function initRealPaymentDate($method, $paydate_one, $paydate_two) {
        $dateOne = new \DateTime($paydate_one);
        $dateTwo = new \DateTime($paydate_two);
        $max = max(array($dateOne, $dateTwo));
        $min = min(array($dateOne, $dateTwo));

        if (array_key_exists($method, $this->_paymentMethods)) {
            $this->_method = $method;
            $datesBasedMin = $this->getDateListMinMax($min->format('Y-m-d'), $max->format('Y-m-d'), '+');
            $datesBasedMax = $this->getDateListMinMax($min->format('Y-m-d'), $max->format('Y-m-d'), '-');

            if (in_array($min->format('Y-m-d'), $datesBasedMax)) {
                $this->_realPaymentDate = $max->format('Y-m-d');
            } else {
                if (in_array($max->format('Y-m-d'), $datesBasedMin)) {
                    $this->_realPaymentDate = $min->format('Y-m-d');
                } else {

                    throw new \Exception("Invalid dates values for " . htmlspecialchars($this->_method) . " method");
                }
            }
        } else {
            $methods = array_keys($this->_paymentMethods);
            throw new Exception('Please enter method from list' . implode(',', $methods));
        }
    }

    /**
     * Generating list of payment dates in given range and given operator
     * if operator is '+' list will be based on minimum if operator is '-' list is based on maximum
     * @param string $min minimum date range format Y-m-d 
     * @param string $max maximum date range format Y-m-d
     * @param string $operator + or - for increasing and decreasing 
     * @return array
     */
    public function getDateListMinMax($min, $max, $operator) {
        $dateList = array();
        $minDate = new \DateTime($min);
        $maxDate = new \DateTime($max);
        $basedOnDate = ($operator == '+' ? $minDate : $maxDate);
        $count = $this->_paymentMethods[$this->_method]['count'];
        $unit = $this->_paymentMethods[$this->_method]['unit'];
        while ($minDate <= $maxDate) {

            $basedOnDate->modify($operator . $count . ' ' . $unit);
            $date = $basedOnDate->format('Y-m-d');
           
            $dateList[] = $this->nearestValidPaymentDate($date);
        }

        return $dateList;
    }
    /**
     * Return nearest valid date for given date
     * @param string $date
     * @return string
     */
    public function nearestValidPaymentDate($date) {
        // not checking holiday after increase because holidays take priority
        if ($this->isWeekend($date)) {
            while ($this->isWeekend($date)) {
                $date = $this->decreaseDate($date, 1);
            }
        }
        // holiday takes priority 
        if ($this->isHoliday($date)) {
            $date = $this->increaseDate($date, 1);
            while (!$this->isValidPaydate($date)) {
                $date = $this->increaseDate($date, 1);
            }
        }
        return $date;
    }

    /**
     * Gives number of day in week
     * @param string $date format Y-m-d
     * @return interger
     */
    public function getDayNumberInWeek($date) {
        return date('N', strtotime($date));
    }

    /**
     * Return list of paymnet methods
     * @return array
     */
    public function getPaymentMethods() {
        return array_keys($this->_paymentMethods);
    }

    public function generateCustomHolidays($string) {
        return explode(',', $string);
    }

}

